/*
 * Copyright (C) 2021 National University of Singapore
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * VERY SIMPLE CONTROL-FLOW-RECOVERY ANALYSIS.
 *
 * This is a very simple control-flow (jump target) recovery analysis.
 *
 * NOTES:
 * - The analysis is not intended to be accurate, and is allowed to
 *   overapproximate the set of basic-blocks.  It is intended for applications
 *   where such overapproximations can be tolerated.
 * - The implementation assumes well-behaved compiler idioms, i.e., jump
 *   tables of a specific form.
 * - The current implementation is somewhat basic and could be improved.
 */

static bool option_cfg_debug = false;
#define DEBUG(targets, target, msg, ...)                                \
    do                                                                  \
    {                                                                   \
        if (option_cfg_debug &&                                         \
                (targets).find(target) == (targets).end())              \
        {                                                               \
            fprintf(stderr, msg "\n", ##__VA_ARGS__);                   \
        }                                                               \
    }                                                                   \
    while (false)

typedef std::map<intptr_t, const Elf64_Rela *> RelaInfo;
enum TargetKind : uint8_t
{
    TARGET_ENTRY,
    TARGET_DIRECT,
    TARGET_INDIRECT
};
typedef std::map<intptr_t, TargetKind> Targets;

/*
 * Insert target information.
 */
static void addTarget(intptr_t target, TargetKind kind, Targets &targets)
{
    auto r = targets.insert({target, kind});
    if (!r.second)
    {
        // Existing entry found:
        switch (r.first->second)
        {
            case TARGET_ENTRY:
                r.first->second = kind;
                break;
            case TARGET_DIRECT:
                if (kind == TARGET_INDIRECT)
                    r.first->second = kind;
                break;
            default:
                break;
        }
    }  
}

/*
 * Get the bounds of a buffer assuming T-aligment.
 */
template <typename T>
static std::pair<const T *, const T *> getBounds(const uint8_t *lb0,
    const uint8_t *ub0)
{
    uintptr_t lb = (uintptr_t)lb0, ub = (uintptr_t)ub0;
    if (lb % sizeof(T) != 0)
    {
        lb += sizeof(T);
        lb -= lb % sizeof(T);
    }
    if (ub % sizeof(T) != 0)
        ub -= ub % sizeof(T);
    return {(const T *)lb, (const T *)ub}; 
}

/*
 * Find the instruction corresponding to the address.  Returns SIZE_MAX if the
 * corresponding instruction is not found.
 */
static size_t findInstr(const Instr *Is, size_t size, intptr_t address)
{
    ssize_t lo = 0, hi = (ssize_t)size-1;
    while (lo <= hi)
    {
        ssize_t mid = (lo + hi) / 2;
        if ((intptr_t)Is[mid].address < address)
            lo = mid+1;
        else if ((intptr_t)Is[mid].address > address)
            hi = mid-1;
        else
            return (size_t)mid;
    }
    return SIZE_MAX;
}

/*
 * Code analysis pass: find all probable code targets.
 */
static void CFGCodeAnalysis(const ELF *elf, bool pic, const Instr *Is,
    size_t size, std::set<intptr_t> &tables, Targets &targets)
{
    // STEP (1): Calculate a rough-cut of the targets:
    intptr_t next = INTPTR_MIN;
    for (size_t i = 0; i < size; i++)
    {
        InstrInfo I0, *I = &I0;
        getInstrInfo(elf, Is + i, I);
        if (next != I->address)
        {
            // [HEURISTIC] This is the first instruction after a "gap" in the
            // executable code.  Thus, something probably jumps here, so is
            // considered a jump target.
            DEBUG(targets, I->address, "Entry : %p", (void *)I->address);
            addTarget(I->address, TARGET_ENTRY, targets);
        }
        next = I->address + I->size;

        intptr_t target = INTPTR_MIN;
        switch (I->mnemonic)
        {
            case MNEMONIC_MOV:
                if (pic || I->op[0].type != OPTYPE_IMM)
                    continue;

                // [HEURISTIC] This instruction may be moving a jump target
                // into another location for later use.  Thus, we consider the
                // immediate value to be a target if it happens to point to a
                // valid instruction.
                target = (intptr_t)I->op[0].imm;
                if (findInstr(Is, size, target) < size)
                {
                    DEBUG(targets, target, "Load  : %p", (void *)target);
                    addTarget(target, TARGET_INDIRECT, targets);
                }
                continue;

            case MNEMONIC_LEA:
                if (I->op[0].type != OPTYPE_MEM ||
                        I->op[0].mem.base != REGISTER_RIP)
                    continue;

                // [HEURISTIC] Similar to the "mov" case but for PIC.
                target = (intptr_t)I->address + (intptr_t)I->size +
                    (intptr_t)I->op[0].mem.disp;
                if (findInstr(Is, size, target) < size)
                {
                    DEBUG(targets, target, "Load  : %p", (void *)target);
                    addTarget(target, TARGET_INDIRECT, targets);
                }
                else if (pic)
                {
                    // This does not point to an instruction, but may be
                    // pointing to the base of a PIC-style jump-table.  We
                    // save the address for later analysis.
                    tables.insert(target);
                }
                continue;

            case MNEMONIC_RET:
                DEBUG(targets, next, "Next  : %p", (void *)next);
                addTarget(next, TARGET_ENTRY, targets);
                continue;
            case MNEMONIC_JMP:
                DEBUG(targets, next, "Next  : %p", (void *)next);
                addTarget(next, TARGET_ENTRY, targets);
                break;
            case MNEMONIC_CALL:
                break;
            case MNEMONIC_JO: case MNEMONIC_JNO: case MNEMONIC_JB:
            case MNEMONIC_JAE: case MNEMONIC_JE: case MNEMONIC_JNE:
            case MNEMONIC_JBE: case MNEMONIC_JA: case MNEMONIC_JS:
            case MNEMONIC_JNS: case MNEMONIC_JP: case MNEMONIC_JNP:
            case MNEMONIC_JL: case MNEMONIC_JGE: case MNEMONIC_JLE:
            case MNEMONIC_JG:
                // The branch-not-taken is considered a jump target:
                DEBUG(targets, next, "NotTkn: %p", (void *)next);
                addTarget(next, TARGET_DIRECT, targets);
                break;
            default:
                continue;
        }

        // If we reach here then the instruction is a jump or call.
        if (I->op[0].type != OPTYPE_IMM)
        {
            // For indirect jumps/call, we do not directly know the target.
            continue;
        }
        target = (intptr_t)I->address + (intptr_t)I->size +
            (intptr_t)I->op[0].imm;
        DEBUG(targets, target, "Target: %p", (void *)target);
        addTarget(target, TARGET_DIRECT, targets);
    }

    // Symbols are assumed to be jump targets:
    for (unsigned i = 0; i < 2; i++)
    {
        const SymbolInfo &syms = (i == 0? getELFDynSymInfo(elf):
                                          getELFSymInfo(elf));
        for (auto &entry: syms)
        {
            const Elf64_Sym *sym = entry.second;
            if (sym->st_shndx == SHN_UNDEF ||
                    ELF64_ST_TYPE(sym->st_info) != STT_FUNC)
                continue;
            intptr_t target = sym->st_value;
            DEBUG(targets, target, "Symbol: %p", (void *)target);
            addTarget(target, TARGET_INDIRECT, targets);
        }
    }
}

/*
 * Section analysis pass: find potential code pointers in data.
 */
static void CFGSectionAnalysis(const ELF *elf, bool pic, const char *name,
    const Elf64_Shdr *shdr, const Instr *Is, size_t size, const RelaInfo relas,
    const std::set<intptr_t> &tables, Targets &targets)
{
    if ((shdr->sh_flags & SHF_EXECINSTR) != 0 || shdr->sh_addr == 0x0)
        return;
    
    const uint8_t *sh_data = getELFData(elf) + shdr->sh_offset;
    size_t sh_size = shdr->sh_size;
    
    if (!pic)
    {
        if (shdr->sh_type != SHT_PROGBITS)
            return;

        // Scan the data for absolute addresses.
        auto bounds = getBounds<intptr_t>(sh_data, sh_data + sh_size);
        for (const intptr_t *p = bounds.first; p < bounds.second; p++)
        {
            intptr_t target = *p;
            if (target != 0 && findInstr(Is, size, target) < size)
            {
                // "Probably" a jump target.
                DEBUG(targets, target, "Data  : %p", (void *)target);
                addTarget(target, TARGET_INDIRECT, targets);
            }
        }
        return;
    }
    else if (pic)
    {
        if (shdr->sh_type == SHT_PROGBITS && (shdr->sh_flags & SHF_WRITE) == 0)
        {
            // Scan the data for PIC-style jump tables.
            auto bounds = getBounds<int32_t>(sh_data, sh_data + sh_size);
            for (const int32_t *p = bounds.first; p < bounds.second; )
            {
                intptr_t table = (intptr_t)shdr->sh_addr +
                    ((intptr_t)p - (intptr_t)sh_data);
                auto i = tables.find(table);
                if (i == tables.end())
                {
                    p++;
                    continue;
                }

                // This is "probably" a PIC-style jump table.
                for (const int32_t *q = p++; q < bounds.second; q++, p = q)
                {
                    intptr_t offset = (intptr_t)*q;
                    intptr_t target = table + offset;
                    if (findInstr(Is, size, target) >= size)
                        break;
                    DEBUG(targets, target, "JmpTbl: %p%+zd = %p",
                        (void *)table, offset, (void *)target);
                    // Jump tables are treated as direct:
                    addTarget(target, TARGET_DIRECT, targets);
                }
            }
        }

        if (shdr->sh_type == SHT_PROGBITS)
        {
            // Scan for code pointers using relocation information.
            auto bounds = getBounds<int64_t>(sh_data, sh_data + sh_size);
            for (const int64_t *p = bounds.first; p < bounds.second; p++)
            {
                intptr_t offset = (intptr_t)shdr->sh_addr +
                    ((intptr_t)p - (intptr_t)sh_data);
                auto i = relas.find(offset);
                if (i == relas.end())
                    continue;
                const Elf64_Rela *rela = i->second;
                
                intptr_t target = *p;
                if (findInstr(Is, size, target) >= size)
                    continue;
                DEBUG(targets, target, "Reloc : %p", (void *)target);
                addTarget(target, TARGET_INDIRECT, targets);
            }
        }
    }
}

/*
 * Data analysis pass: find potential code pointers in data.
 */
static void CFGDataAnalysis(const ELF *elf, bool pic, const Instr *Is,
    size_t size, const std::set<intptr_t> &tables, Targets &targets)
{
    // Gather relocation information:
    const SectionInfo &sections = getELFSectionInfo(elf);
    RelaInfo relas;
    for (const auto &entry: sections)
    {
        const Elf64_Shdr *shdr = entry.second;
        if (shdr->sh_type != SHT_RELA)
            continue;
        const uint8_t *sh_data = getELFData(elf) + shdr->sh_offset;
        size_t sh_size = shdr->sh_size;
        const Elf64_Rela *rela = (const Elf64_Rela *)sh_data;
        const Elf64_Rela *rela_end = rela + sh_size / sizeof(Elf64_Rela);
        for (; rela < rela_end; rela++)
        {
            if (ELF64_R_TYPE(rela->r_info) == R_X86_64_RELATIVE)
                relas.insert({rela->r_offset, rela});
        }
    }

    // Analyze each data section:
    for (const auto &entry: sections)
        CFGSectionAnalysis(elf, pic, entry.first, entry.second, Is, size,
            relas, tables, targets);
}

/*
 * Analyze the given ELF binary for potential jump targets.
 */
static void CFGAnalysis(const ELF *elf, const Instr *Is, size_t size,
    Targets &targets)
{
    bool pic = false;
    switch (getELFType(elf))
    {
        case ELFTYPE_DSO: case ELFTYPE_PIE:
            pic = true;
            break;
        default:
            break;
    }

    // Pass #1: Find all code targets. 
    std::set<intptr_t> tables;
    CFGCodeAnalysis(elf, pic, Is, size, tables, targets);
    
    // Pass #2: Find all data targets.
    CFGDataAnalysis(elf, pic, Is, size, tables, targets);

    // Pass #3: "Clean up" the targets.
    Targets new_targets;
    for (const auto &entry: targets)
    {
        intptr_t target = entry.first;
        TargetKind kind = entry.second;

        // Find the corresponding instruction:
        size_t i = findInstr(Is, size, target);
        if (i >= size)
            continue;

        // Skip any NOPs/INTs/UDs:
        InstrInfo I0, *I = &I0;
        for (; i < size; i++)
        {
            getInstrInfo(elf, Is + i, I);
            bool stop = false;
            switch (I->mnemonic)
            {
                case MNEMONIC_NOP:
                case MNEMONIC_INT: case MNEMONIC_INT1: case MNEMONIC_INT3:
                case MNEMONIC_INTO:
                case MNEMONIC_UD0: case MNEMONIC_UD1: case MNEMONIC_UD2:
                    break;
                default:
                    stop = true;
                    break;
            }
            if (stop)
                break;
            if (i+1 < size &&
                    (intptr_t)Is[i+1].address != I->address + I->size)
                i = SIZE_MAX;
        }
        if (i >= size)
            continue;   // No target found.
        
        // Add target:
        addTarget((intptr_t)Is[i].address, kind, new_targets);
    }
    targets.swap(new_targets);
}

